<?php

declare(strict_types=1);

namespace Drupal\addevent\Plugin\Block;

use Drupal\Core\Annotation\ContextDefinition;
use Drupal\Core\Annotation\Translation;
use Drupal\Core\Block\Annotation\Block;
use Drupal\Core\Block\BlockBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Utility\Token;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * @Block(
 *   id = "addevent_calendar_button",
 *   admin_label = @Translation("Add To Calendar"),
 *   category = "AddEvent",
 *   context_definitions = {
 *     "node" = @ContextDefinition("entity:node", label = @Translation("Node"))
 *   }
 * )
 */
class AddToCalendarBlock extends BlockBase implements ContainerFactoryPluginInterface {

  /**
   * @var \Drupal\Core\Utility\Token
   */
  protected $token;

  /**
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * {@inheritDoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    Token $token,
    ModuleHandlerInterface $module_handler
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition
    );
    $this->token = $token;
    $this->moduleHandler = $module_handler;
  }

  /**
   * {@inheritDoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('token'),
      $container->get('module_handler')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function defaultConfiguration(): array {
    return [
        'settings' => [
          'button' => $this->t('Add to Calendar'),
          'event' => [
            'title' => NULL,
            'end' => NULL,
            'start' => NULL,
            'description' => NULL,
            'timezone' => date_default_timezone_get(),
            'location' => NULL,
            'organizer' => NULL,
            'organizer_email' => NULL,
            'facebook_event' => NULL,
            'all_day_event' => NULL,
            'date_format' => 'MM/DD/YYYY',
            'alarm_reminder' => NULL,
            'recurring' => NULL,
            'calname' => NULL,
            'attendees' => NULL,
            'client' => NULL,
            'status' => 'CONFIRMED',
            'method' => NULL,
            'transp' => 'OPAQUE'
          ],
          'customization' => [
            'css' => TRUE,
            'mouse' => FALSE,
            'license' => NULL,
            'providers' => [],
          ],
          'data_attributes' => [
            'data-direct' => NULL,
            'data-dropdown-x' => NULL,
            'data-dropdown-y' => NULL,
            'data-render' => NULL,
            'data-intel' => NULL,
            'data-google-api' => NULL,
            'data-outlook-api' => NULL,
          ],
        ]] + parent::defaultConfiguration();
  }

  /**
   * {@inheritDoc}
   */
  public function blockForm($form, FormStateInterface $form_state): array {
    $form = parent::blockForm($form, $form_state);

    $settings = $this->getAddEventSettings();

    $form['settings'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('AddEvent Settings'),
      '#tree' => TRUE,
    ];
    $form['settings']['button'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Button Text'),
      '#required' => TRUE,
      '#default_value' => $settings['button']
    ];
    $form['settings']['event'] = [
      '#type' => 'details',
      '#title' => $this->t('Event Information'),
      '#description' => $this->t('Input a token or string value which will be
        used to create the calendar button.'),
      '#open' => TRUE,
    ];

    if ($this->moduleHandler->moduleExists('token')) {
      $form['settings']['event']['tokens'] = [
        '#theme' => 'token_tree_link',
        '#token_types' => ['node'],
      ];
    }
    $eventConfig = $settings['event'] ?? [];

    $form['settings']['event']['title'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Title'),
      '#required' => TRUE,
      '#description' => $this->t('Input the title of the event.'),
      '#default_value' => $eventConfig['title'],
    ];
    $form['settings']['event']['description'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Description'),
      '#description' => $this->t('Input the description of the event.'),
      '#default_value' => $eventConfig['description'],
    ];
    $form['settings']['event']['start'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Start'),
      '#required' => TRUE,
      '#description' => $this->t('Input the start date of the event.'),
      '#default_value' => $eventConfig['start'],
    ];
    $form['settings']['event']['end'] = [
      '#type' => 'textfield',
      '#title' => $this->t('End'),
      '#description' => $this->t('Input the end date of the event. If end is not
        defined, the end date is automatically set to start date plus one hour.'),
      '#default_value' => $eventConfig['end'],
    ];
    $form['settings']['event']['timezone'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Timezone'),
      '#description' => $this->t('Input the event time zone.'),
      '#default_value' => $eventConfig['timezone'],
    ];
    $form['settings']['event']['location'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Location'),
      '#description' => $this->t('Input the event location.'),
      '#default_value' => $eventConfig['location'],
    ];
    $form['settings']['event']['organizer'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Organizer'),
      '#description' => $this->t('Input the event organizer.'),
      '#default_value' => $eventConfig['organizer'],
    ];
    $form['settings']['event']['organizer_email'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Organizer Email'),
      '#description' => $this->t('Input the event organizer e-mail. Required if
        the organizer field has a value!'),
      '#states' => [
        'visible' => [
          ':input[name="settings[settings][event][organizer]"]' => [
            'filled' => TRUE
          ]
        ]
      ],
      '#default_value' => $eventConfig['organizer_email'],
    ];
    $form['settings']['event']['facebook_event'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Facebook Event'),
      '#description' => $this->t('Input the event Facebook link.'),
      '#default_value' => $eventConfig['facebook_event'],
    ];
    $form['settings']['event']['all_day_event'] = [
      '#type' => 'textfield',
      '#title' => $this->t('All Day'),
      '#description' => $this->t('Input true/false if the the event is all day.'),
      '#default_value' => $eventConfig['all_day_event'],
    ];
    $form['settings']['event']['date_format'] = [
      '#type' => 'textfield',
      '#required' => TRUE,
      '#title' => $this->t('Date Format'),
      '#description' => $this->t('Input the format of the date in start and end.
        Accepts the values MM/DD/YYYY or DD/MM/YYYY only.'),
      '#default_value' => $eventConfig['date_format'],
    ];
    $form['settings']['event']['alarm_reminder'] = [
      '#type' => 'number',
      '#title' => $this->t('Alarm Reminder'),
      '#description' => $this->t('Set event reminder. Trigger an event reminder
        e.g. "15" minutes before the event starts.'),
      '#default_value' => $eventConfig['alarm_reminder'],
    ];
    $form['settings']['event']['recurring'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Recurring'),
      '#description' => $this->t('Input the event recurring rules. Use
        <a href="@url" target="_blank">TextMagic RRule Generator</a>.', [
          '@url' => 'https://www.textmagic.com/free-tools/rrule-generator'
      ]),
      '#default_value' => $eventConfig['recurring'],
    ];
    $form['settings']['event']['calname'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Calendar Name'),
      '#description' => $this->t('Input the custom file naming of the .ics file
        used for e.g. Outlook and Apple Calendar. <br/><strong>Note:</strong> If
        left empty the event title is used.'),
      '#default_value' => $eventConfig['calname'],
    ];
    $form['settings']['event']['attendees'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Attendees'),
      '#description' => $this->t('Input one or multiple attendees (pre-populated)
        to the event. Separate multiple email addresses with a comma.'),
      '#default_value' => $eventConfig['attendees'],
    ];
    $form['settings']['event']['status'] = [
      '#type' => 'select',
      '#title' => $this->t('Status'),
      '#description' => $this->t('Define the status parameter in the .ics file
        used for e.g. Outlook and Apple Calendar.'),
      '#options' => [
        'COMPLETED' => $this->t('Completed'),
        'CONFIRMED' => $this->t('Confirmed'),
        'TENTATIVE' => $this->t('Tentative'),
        'CANCELLED' => $this->t('Cancelled'),
        'IN-PROCESS' => $this->t('In Progress'),
        'NEEDS-ACTION' => $this->t('Need Action'),
      ],
      '#default_value' => $eventConfig['status'],
    ];
    $form['settings']['event']['client'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Client'),
      '#description' => $this->t('The client field refers to your account ID.'),
      '#default_value' => $eventConfig['client'],
    ];
    $form['settings']['event']['method'] = [
      '#type' => 'select',
      '#title' => $this->t('Method'),
      '#description' => $this->t('If defined, a method parameter is added in the
        .ics file used by e.g. Outlook and Apple Calendar..'),
      '#options' => [
        'PUBLISH' => $this->t('Publish'),
        'REQUEST' => $this->t('Request'),
        'CANCEL' => $this->t('Cancel'),
        'REFRESH' => $this->t('Refresh'),
      ],
      '#empty_option' => $this->t('- None -'),
      '#default_value' => $eventConfig['method'],
    ];
    $form['settings']['event']['transp'] = [
      '#type' => 'select',
      '#title' => $this->t('Transparent'),
      '#description' => $this->t('Determines whether the event appears as "free"
        or "busy" on the users calendar.'),
      '#options' => [
        'OPAQUE' => $this->t('Busy'),
        'TRANSPARENT' => $this->t('Free'),
      ],
      '#default_value' => $eventConfig['transp'],
    ];

    $form['settings']['data_attributes'] = [
      '#type' => 'details',
      '#title' => $this->t('Data Attributes'),
      '#open' => FALSE,
    ];
    $dataAttributesConfig = $settings['data_attributes'] ?? [];

    $form['settings']['data_attributes']['data-direct'] = [
      '#type' => 'select',
      '#title' => $this->t('Direct'),
      '#description' => $this->t('The attribute disables the dropdown and makes
        the button a direct link to the service specified.'),
      '#options' => $this->getCalendarProviderOptions(),
      '#empty_option' => $this->t('- Default -'),
      '#default_value' => $dataAttributesConfig['data-direct'],
    ];
    $form['settings']['data_attributes']['data-dropdown-x'] = [
      '#type' => 'select',
      '#title' => $this->t('Dropdown x-axis'),
      '#description' => $this->t(
        'Control the horizontal position of the dropdown.'
      ),
      '#options' => [
        'auto' => $this->t('Auto'),
        'left' => $this->t('Left'),
        'right' => $this->t('Right'),
      ],
      '#empty_option' => $this->t('- Default -'),
      '#default_value' => $dataAttributesConfig['data-dropdown-x'],
    ];
    $form['settings']['data_attributes']['data-dropdown-y'] = [
      '#type' => 'select',
      '#title' => $this->t('Dropdown y-axis'),
      '#description' => $this->t(
        'Control the vertical position of the dropdown.'
      ),
      '#options' => [
        'auto' => $this->t('Auto'),
        'up' => $this->t('Up'),
        'down' => $this->t('Down'),
      ],
      '#empty_option' => $this->t('- Default -'),
      '#default_value' => $dataAttributesConfig['data-dropdown-y'],
    ];
    $form['settings']['data_attributes']['data-render'] = [
      '#type' => 'select',
      '#title' => $this->t('Render'),
      '#description' => $this->t(
        'The attribute ensures that the options are generated and displayed at
        load. The attribute furthermore disables the default CSS.'
      ),
      '#options' => [
        'inline-buttons' => $this->t('Inline Button'),
      ],
      '#empty_option' => $this->t('- Default -'),
      '#default_value' => $dataAttributesConfig['data-render'],
    ];
    $form['settings']['data_attributes']['data-intel'] = [
      '#type' => 'select',
      '#title' => $this->t('Intel'),
      '#description' => $this->t(
        'If we can detect which device the user is on (e.g. an iDevice) it makes
        sense to automatically select the "Apple" option on click instead of
        showing the dropdown option.'
      ),
      '#options' => [
        'true' => $this->t('True'),
        'false' => $this->t('False'),
      ],
      '#empty_option' => $this->t('- Default -'),
      '#default_value' => $dataAttributesConfig['data-intel'],
    ];
    $form['settings']['data_attributes']['data-google-api'] = [
      '#type' => 'select',
      '#title' => $this->t('Google API'),
      '#description' => $this->t(
        'The attribute enables the Google Events API for adding events to users
        calendar.'
      ),
      '#options' => [
        'true' => $this->t('True'),
      ],
      '#empty_option' => $this->t('- Default -'),
      '#default_value' => $dataAttributesConfig['data-google-api'],
    ];
    $form['settings']['data_attributes']['data-outlook-api'] = [
      '#type' => 'select',
      '#title' => $this->t('Outlook API'),
      '#description' => $this->t(
        'The attribute enables the Microsoft Events API for adding events to
        users calendar.'
      ),
      '#options' => [
        'true' => $this->t('True'),
      ],
      '#empty_option' => $this->t('- Default -'),
      '#default_value' => $dataAttributesConfig['data-outlook-api'],
    ];

    $form['settings']['customization'] = [
      '#type' => 'details',
      '#title' => $this->t('Customization'),
      '#open' => FALSE,
    ];
    $customizationConfig = $settings['customization'] ?? [];

    $form['settings']['customization']['mouse'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Mouse'),
      '#description' => $this->t('Show options dropdown on mouseover.'),
      '#default_value' => $customizationConfig['mouse'],
    ];
    $form['settings']['customization']['css'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('CSS'),
      '#description' => $this->t('Use the default add to calendar styles.'),
      '#default_value' => $customizationConfig['css'],
    ];
    $form['settings']['customization']['license'] = [
      '#type' => 'textfield',
      '#title' => $this->t('License'),
      '#description' => $this->t('Input the client account ID. Removes AddEvent
        branding.'),
      '#default_value' => $customizationConfig['license'],
    ];
    $form['settings']['customization']['providers'] = [
      '#type' => 'table',
      '#header' => [
        $this->t('Provider'),
        $this->t('Show'),
        $this->t('Text'),
        $this->t('Weight'),
      ],
      '#tabledrag' => [
        [
          'action' => 'order',
          'relationship' => 'sibling',
          'group' => 'table-sort-weight',
        ],
      ],
    ];
    $calendarProviders = array_replace_recursive(
      $this->getCalendarProviders(),
      $customizationConfig['providers'] ?? []
    );
    $providers = &$form['settings']['customization']['providers'];

    foreach ($this->sortCalendarProviders($calendarProviders) as $name => $provider) {
      if (!isset($provider['label'])) {
        continue;
      }
      $providerLabel = $provider['label'];
      $providerWeight = $provider['weight'] ?? 0;

      $providers[$name]['#weight'] = $providerWeight;
      $providers[$name]['#attributes']['class'][] = 'draggable';

      $providers[$name]['provider'] = [
        '#plain_text' => $providerLabel,
      ];
      $providers[$name]['show'] = [
        '#type' => 'checkbox',
        '#default_value' => $provider['show'],
      ];
      $providers[$name]['text'] = [
        '#type' => 'textfield',
        '#required' => TRUE,
        '#default_value' => $provider['text']
      ];
      $providers[$name]['weight'] = [
        '#type' => 'weight',
        '#title' => $this->t('Weight for @label', [
          '@label' => $providerLabel,
        ]),
        '#title_display' => 'invisible',
        '#default_value' => $providerWeight,
        '#attributes' => [
          'class' => [
            'table-sort-weight',
          ],
        ],
      ];
    }

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function blockValidate($form, FormStateInterface $form_state): void {
    parent::blockValidate($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state): void {
    parent::blockSubmit($form, $form_state);
    $values = $form_state->getValues();

    $this->setConfigurationValue(
      'settings',
      $values['settings']
    );
  }

  /**
   * {@inheritDoc}
   */
  public function build(): array {
    $settings = $this->getAddEventSettings();

    return [
      '#theme' => 'add_to_calendar_button',
      '#button_text' => $settings['button'],
      '#parameters' => $this->buildEventParameters(),
      '#attached' => [
        'library' => 'addevent/add-to-calendar',
        'drupalSettings' => [
          'addevent' => [
            'settings' => $this->buildEventCustomizationSettings()
          ]
        ]
      ],
      '#data_attributes' => array_filter($settings['data_attributes']),
    ];
  }

  /**
   * Get the AddEvent settings.
   *
   * @return array
   *   An array of the AddEvent settings.
   */
  protected function getAddEventSettings(): array {
    return $this->getConfiguration()['settings'];
  }

  /**
   * Build the event customization settings.
   *
   * @return array
   *   An array of formatted customization settings.
   */
  protected function buildEventCustomizationSettings(): array {
    $order = [];
    $settings = [];
    $customization = $this->getAddEventSettings()['customization'];

    foreach ($customization['providers'] as $name => $provider) {
      if (!isset($provider['show']) || !isset($provider['text'])) {
        continue;
      }

      if ($show = $provider['show']) {
        $order[] = $name;
        $settings[$name] = [
          'show' => (bool) $show,
          'text' => (string) $provider['text']
        ];
      }
    }

    return [
      'css' => (bool) $customization['css'],
      'mouse' => (bool) $customization['mouse'],
      'license' => (string) $customization['license'],
      'dropdown' => [
        'order' => (string) implode(',', array_values($order))
      ]
    ] + $settings;
  }

  /**
   * Build the tokenized event parameters.
   *
   * @return array
   *   An array of event parameters.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  protected function buildEventParameters(): array {
    $variables = array_filter(
      $this->getAddEventSettings()['event']
    );
    $node = $this->getBlockNodeContext();

    foreach ($variables as &$variable) {
      $variable = $this->token->replace(
        $variable,
        ['node' => $node],
        ['clear' => TRUE]
      );
    }

    return $variables;
  }

  /**
   * Get the calendar provider options.
   *
   * @return array
   *   An array of the provider options.
   */
  protected function getCalendarProviderOptions(): array {
    $options = [];

    foreach ($this->getCalendarProviders() as $name => $provider) {
      if (!isset($provider['label'])) {
        continue;
      }
      $options[$name] = $provider['label'];
    }

    return $options;
  }

  /**
   * Get the calendar providers information.
   *
   * @return array
   *   An array of the calendar providers.
   */
  protected function getCalendarProviders(): array {
    return [
      'appleical' => [
        'label' => $this->t('Apple'),
        'show' => TRUE,
        'weight' => 0
      ],
      'google' => [
        'show' => TRUE,
        'label' => $this->t('Google'),
        'text' => $this->t('Google <em>(online)</em>'),
        'weight' => 0
      ],
      'office365' => [
        'show' => 1,
        'label' => $this->t('Office 365'),
        'text' => $this->t('Office 365 <em>(online)</em>'),
        'weight' => 0
      ],
      'outlook' => [
        'show' => TRUE,
        'label' => $this->t('Outlook'),
        'weight' => 0
      ],
      'outlookcom' => [
        'show' => TRUE,
        'label' => $this->t('Outlook.com'),
        'text' => $this->t('Outlook.com <em>(online)</em>'),
        'weight' => 0
      ],
      'yahoo' => [
        'show' => FALSE,
        'label' => $this->t('Yahoo'),
        'text' => $this->t('Yahoo <em>(online)</em>'),
        'weight' => 0
      ],
      'facebook' => [
        'show' => FALSE,
        'label' => $this->t('Facebook'),
        'weight' => 0
      ],
    ];
  }

  /**
   * Sort the calendar providers.
   *
   * @param array $providers
   *   An array of calendar providers.
   *
   * @return array
   *   An array of providers sorted by the weight value.
   */
  protected function sortCalendarProviders(array $providers): array {
    uasort($providers, function ($a, $b) {
      if ($a['weight'] == $b['weight']) {
        return 0;
      }
      return $a['weight'] > $b['weight'] ? 1 : -1;
    });

    return $providers;
  }

  /**
   * Get the node entity from the context.
   *
   * @return EntityInterface
   *   The Drupal entity instance.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  protected function getBlockNodeContext(): EntityInterface {
    $context = $this->getContext('node');

    if (!$context->hasContextValue()) {
      throw new \Exception(
        'Unable to retrieve the node context value.'
      );
    }

    return $context->getContextData()->getValue();
  }
}
